{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}
{$longstrings on}
{$codepage utf-8}

unit cl_framework_winapi;

interface

uses SysUtils, classes, Windows
    , cl_windows_missing
    //, jwawinuser
    , math, cge, cl_gamepad, cl_wintab;

  {$include cl_winman.h}

type
  TWinApiFramework = class (TAbstractFramework)
    //Please note that this framework is a half-measure. I slapped it on top of
    //  old WindowManager classes, so for Win32/64 and X11-32 many methods
    //  are just wrappers. Tooo much to rehaul otherwise.
    //But any new frameworks, like gtk2, will be self-contained, without any
    //  window managers whatsoever.
  protected
    BatteryStatusCheckMoment: TdateTime;
    PrevBatteryRemaining: integer;
  public
    WindowManager: TWindowManager;
    WinTabManager: TWinTabmanager;
    GamepadManager: TGamepadManager;
    SoundManager: TMotherSoundMan;
    constructor Create; override;
    destructor Destroy; override;
    procedure SetPointerImage(p: pointer; xhs, yhs: integer); override;
    procedure SwapBuffers; override; //calls WindowManager.SwapBuffers
    function MouseX: float; override;
    function MouseY: float; override;
    Procedure MainThreadLoop; override;
    procedure SetVolume(volume: float); override;
    procedure NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); override;
    procedure PlaySound(samples, rate: integer; data: pointer); override;
    procedure PlayWav(filename: TFileNameString); override;
    procedure CheckBatteryStatus; override;
    procedure SetWindowName(n: PUnicodeChar); override;
  end;


implementation

 //the old window manager basis, mostly platform-dependent
 {$include cl_winman.inc}

 //the higher, platform-independent window manager logic. Still obsolete.
 {$include cl_window.inc}


  constructor TWinApiFramework.Create;
  begin
    inherited;

    Cge.Framework:= Self; //other classes need to access it.

    if Mother^.Debug.Enabled then AddLog('Starting the window manager...');

    PrevBatteryRemaining:= -1;

    WindowManager:= TWindowManager.Create;
    EngineObject.AddObj(WindowManager);
    WindowManager.pointer_to_instance:= pointer(@Self.WindowManager);

    //CgeEmptyCursor;
    Self.SetPointerImage(nil, 0, 0);

    Console.InitGL;

    if Mother^.Debug.Enabled then AddLog('Starting the pen tablet manager...');
    WinTabManager:= TWinTabmanager.Create;
    EngineObject.AddObj(WinTabManager);

    GamepadManager:= TGamepadManager.Create;
    GamepadManager.pointer_to_instance:= pointer(@Self.GamepadManager);
    EngineObject.AddObj(GamepadManager);

    Mother^.Display.BeginBlackout(INITIAL_BLACKOUT);

    Console.DrawBackground;
    Console.Draw;

    if Mother^.Debug.Enabled then AddLog('Loading the fixed font...');
    InitFixedFont;

    if Mother^.Debug.Enabled then AddLog('Loading the stub texture...');
    InitStubTexture;

    LoadQF;

    if Mother^.Debug.Enabled then AddLog('Initializing sound...');

    try
      if Config.Bool[Mother^.State.OstID, 'use_directsound_not_openal']
        then SoundManager:= TDirectXSoundMan.Create
        else SoundManager:= TOpenALSoundman.Create;
      EngineObject.AddObj(SoundManager);
    except
      pointer(SoundManager):= nil;
      if Config.bool['sound', 'allow_running_without']
      then
        AddWarning (MI_CAPTION_NOSOUND, StopDying(),[])
      else
        AddStopMessage(StopDying()); //clear the error messages buffer by copying them into a error screen.
    end;


  end;

  destructor TWinApiFramework.destroy;
  begin
    pointer(Cge.Framework) := nil;
    inherited;
  end;


  Procedure TWinApiFramework.MainThreadLoop;
  begin
    try
      Mother^.State.ProgramInitialized:= Yes;

      WindowManager.MainLoop;

      if Mother^.Debug.Enabled then AddLog('...exited the main loop.');
    Except
      EverythingIsOk:=No;
      if not (ExceptObject is EDying) then Try
          Die(MI_ERROR_UNCAUGHT_EXCEPTION);
      Except
      End;
    End;
  end;


  procedure TWinApiFramework.CheckBatteryStatus;
  //needed this for limiting rendering settings when laptop runs on battery
  var
    ps: SYSTEM_POWER_STATUS;
//    F: TextFile;
    s: TFileStream;
    u: Utf8String;
  begin
    if Now() < BatteryStatusCheckMoment then Exit; //check every 10 seconds
    if not GetSystemPowerStatus(ps) then begin
      if Mother^.State.BatteryRemaining <> 255
        then AddLog(RuEn(
          'Не удалось получить от ОС состояние электропитания. Заряд батареи НЕ учитывается!',
          'Failed to get power status from the OS. The battery charge is NOT accounted for!'));
      Mother^.State.BatteryRemaining:= 255;
      BatteryStatusCheckMoment:= Now() + (5 * 60) / SecondsPerDay; //check again five minutes later
    end
    else if ps.ACLineStatus > 0 then begin
      BatteryStatusCheckMoment:= Now() + 10 / SecondsPerDay; //check often: what if it gets unplugged?
      Mother^.State.BatteryRemaining:= 255;
    end
    else begin
      Mother^.State.BatteryRemaining:= ps.BatteryLifePercent; //it's an unsigned byte, LOL.
      BatteryStatusCheckMoment:= Now() + 30 / SecondsPerDay;  //check two times a minute

      if Mother^.State.BatteryRemaining < 10 then begin
        MakeSurePathExists(UnmangleFileName('*H/logs'));
        s:= TFileStream.Create(UnmangleFileName('*H/logs/last-last-will.txt'), fmCreate);
//        AssignFile(F, _UnmangleFileName('*H/logs/last-last-will.txt'));
//        Rewrite(F);
//        WriteLn(F,
          u:= Utf8Encode(RuEn(
          'Аварийный выход: батарея ноутбука щас сдохнет!',
          'Emergency shutdown: laptop battery is dying!')) + #10#13;
          //);
        s.Write(u[1], Length(u));
        s.Free;
//        CloseFile(F);
        RestartMyself('--last-will');
      end;
    end;

    if (PrevBatteryRemaining <> Mother^.State.BatteryRemaining)
    then begin
      if Mother^.State.BatteryRemaining > 100
        then AddLog(RuEn('Питание от сети.', 'AC power.'))
        else AddLog(RuEn('Заряд батареи %0%.', 'Battery charge %0%.')
                    , [Mother^.State.BatteryRemaining]);
        PrevBatteryRemaining:= Mother^.State.BatteryRemaining;
    end;

    //Disable higher-resolution waits for better battery life
    if Mother^.State.BatteryRemaining <= 100 then begin
      if Mother^.Timer.ActiveTBP1 then begin
        if Mother^.Debug.Verbose then AddLog('Calling timeEndPeriod(1)...') ;
        if timeEndPeriod(1) = TIMERR_NOERROR then begin
          Mother^.Timer.ActiveTBP1:= false;
          if Mother^.Debug.Verbose then AddLogOk
        end else
          if Mother^.Debug.Verbose then AddLog('FAILED');
      end;
    end else begin
      if not Mother^.Timer.ActiveTBP1 then begin
        if Mother^.Debug.Verbose then AddLog('Calling timeBeginPeriod(1)...') ;
        if timeBeginPeriod(1) = TIMERR_NOERROR then begin
          Mother^.Timer.ActiveTBP1:= true;
          if Mother^.Debug.Verbose then AddLogOk
        end else
          if Mother^.Debug.Verbose then AddLog('FAILED');
      end;
    end;
   end;

  //must point to a 32x32 ABGR pixels array.
  //NIL causes to hide the pointer.
  procedure TWinApiFramework.SetPointerImage(p: pointer; xhs, yhs: integer);
  begin
    if Assigned(WindowManager) then WindowManager.SetPointerImage(p, xhs, yhs);
  end;

  // performs OpenGL's SwapBuffers for the main framebuffer, nothing more.
  procedure TWinApiFramework.SwapBuffers;
  begin
    if Assigned(WindowManager) then WindowManager.SwapBuffers;
  end;

  function TWinApiFramework.MouseX: float;
  begin
    Result:= 0;
    if Assigned(WindowManager) then Result:= WindowManager.MouseX;
  end;


  function TWinApiFramework.MouseY: float;
  begin
    Result:= 0;
    if Assigned(WindowManager) then Result:= WindowManager.MouseY;
  end;

  procedure TWinApiFramework.SetVolume(volume: float);
  begin
    if Assigned(SoundManager) then SoundManager.setVolume(volume);
  end;


  procedure TWinApiFramework.NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer);
  begin
    if Assigned(SoundManager) then SoundManager.NewData(begins, samples, rate, data);
  end;


  procedure TWinApiFramework.PlaySound(samples, rate: integer; data: pointer);
  begin
    if Assigned(SoundManager) then SoundManager.PlaySound(samples, rate, data);
  end;


  procedure TWinApiFramework.PlayWav(filename: TFileNameString);
  begin
    if Assigned(SoundManager) then SoundManager.PlayWav(filename);
  end;

  procedure TWinApiFramework.SetWindowName(n: PUnicodeChar);
  begin
    if
      {$ifdef cpu32}
        not Assigned(SetWindowTextW) or Mother^.State.Windows98.DoesNotSupportUnicode
      {$else}
        false
      {$endif}
    then begin
      SetWindowText(
        Mother^.Display.WindowHandle,
        PChar(WideToAnsi(n))
      );
    end
    else begin
      SetWindowTextW(
        Mother^.Display.WindowHandle,
        n
      );
    end;
   end;


initialization
  TAbstractFramework.RegisterFramework(TWinApiFramework, 'WinAPI');
finalization

end.

