{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file is contains the basic functions exported from
    the mother module to the game module DLL.

 **********************************************************************}

{$include cl_mother_api_timer.inc}

{$include cl_mother_api_imaging.inc}

{$include cl_mother_api_cursor.inc}


  procedure CgeNewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); cdecl;
  begin
   try
    if Assigned(Framework)
      then Framework.NewSoundData(begins, samples, rate, data);
   except
    ProcessGuardedException();
   end;
  end;

  procedure CgeSetSoundVolume(volume: float); cdecl;
  begin
   try
     if Assigned(Framework)
      then Framework.SetVolume(volume);
   except
    ProcessGuardedException();
   end;
  end;

  procedure _AllocateThreadExceptionState(thread_title: PUnicodeChar); cdecl;
  var
    pes: PMotherSehState;
	  ssaddr: ptruint;
  begin
   //no try block. If *this* simplest thing crashes it means things are already FUBAR.
    pes:= GetExceptionState();

//    FillChar(pes^, sizeof(pes^), 0);
//it comes *already* zeroed out and with ThreadId set. Cheb no baka.

    pes^.IsModuleThread:= true;
    if Assigned(thread_title) then pes^.ThreadTitle:= PUnicodeCharToUnicodeString(thread_title);
  end;

  function _ModuleStopDying(pes: PMotherSehState = nil): PUnicodeChar; cdecl;
  begin
    //again, no try block here
    if not Assigned(pes)
      then pes:= GetExceptionState();
    pes^._StopDyingString:= StopDying(pes);
//bgSay('* _ModuleStopDying pes='+ IntToHex(dword(pes),8)+ ' string='+ pes^._StopDyingString );
    Result:= PUnicodeChar(pes^._StopDyingString);
  end;


  procedure _cgeSetBackground(FileName: PAnsiChar); cdecl;
  begin
    try
    if not Mother^.Display.CanRender or not Assigned(Console) then Exit;
    Console.SetBg(PCharToString(FileName));
   except
    ProcessGuardedException();
   end;
  end;
  
  procedure _cgeSetBackgroundPtr(p: pointer; size: integer); cdecl;
  begin
    try
    if not Mother^.Display.CanRender or not Assigned(Console) then Exit;
    Console.SetBgPtr(p, size);
   except
    ProcessGuardedException();
   end;
  end;

  procedure BeginBlackout(delta: double = 0.0); cdecl;
  begin
    if Assigned(Console) then try Console.BeginBlackout(delta) except end;
  end;

//! Deprecated due to change in paradigm
  {var cgehbPrevNow: TDateTime = 0;

  procedure _cgeHeartBeat(); cdecl;
  var n: TDateTime;
  begin
    try
 //   if Mother^.Module.CurrentlyExecutingInd <> NOT_A_MODULE then Exit;
    if not Mother^.Display.CanRender then Exit;
    n:=(Now() - cgehbPrevNow) * 86400000.0;
    if n < 100 then Exit;
    Console.Draw(false, true);
    _cgeSwapBuffers;
    cgehbPrevNow:=Now;
   except
    ProcessGuardedException();
   end;
  end;
}
  
  function GetWindowExists(): boolean; cdecl;
  begin
    Result:=Assigned(Framework) and (Mother^.Display.WindowExists);
  end;


  procedure _CgeSwapBuffers(); cdecl;
  begin
    try
    Framework.SwapBuffers;
   except
    ProcessGuardedException();
   end;
  end;
  

 var
    PrevTimeStamp: qword = 0;
    
  function GetTimeStamp: qword;
  begin
    asm
      pushf
      {$ifdef cpu64}
        push rdx
        push rax
      {$else}
        push edx
        push eax
      {$endif}
      rdtsc
      mov dword[Result], eax
      mov dword[Result + 4], edx
      {$ifdef cpu64}
        pop rax
        pop rdx
      {$else}
        pop eax
        pop edx
      {$endif}
      popf
    end;
    if Result <= PrevTimeStamp then begin
      Result:=PrevTimeStamp + 1;
      inc(PrevTimeStamp);
    end
    else
      PrevTimeStamp:=Result;
  end;

{  procedure NewSessionID(); cdecl;
  begin
    TimeStamp:=GetTimeStamp();
  End;}

 procedure _SetConfigInt (section, id: PAnsiChar; Value: integer); cdecl;
 begin
   try
   Config[section, id]:=value;
   except
    ProcessGuardedException();
   end;
 end;
 
 function _GetConfigInt (section, id: PAnsiChar): integer; cdecl;
 begin
   try
   Result:=Config[section, id];
   except
    ProcessGuardedException();
   end;
 end;

procedure _SetConfigStr (section, id, Value: PAnsiChar); cdecl;
begin
  try
   Config.Str[section, id]:= Value;
  except
   ProcessGuardedException();
  end;
end;

threadvar sGetConfigStrResult: AnsiString;

function _GetConfigStr (section, id: PAnsiChar): PAnsiChar; cdecl;
begin
  try
   sGetConfigStrResult:= Config.Str[section, id];
   Result:= PAnsiChar(sGetConfigStrResult);
  except
   ProcessGuardedException();
  end;
end;

 function _GetConfigIntCh (section, id: PAnsiChar; min, max:integer): integer; cdecl;
 begin
   try
   Result:=Config.IntChk[section, id, min, max];
   except
    ProcessGuardedException();
   end;
 end;

 procedure _PlaySound(Wavfilename: PAnsiChar); cdecl;
  begin
    try
      if Assigned(Framework) then Framework.PlayWav(Wavfilename);
    except
     ProcessGuardedException();
    end;
  end;

  Procedure _HostLogW(txt: PUnicodeChar); cdecl;
  begin
//    try
    AddLog(PUnicodeCharToUnicodeString(txt));
//    LeaveGuardedSection
  end;
  
  Procedure DbgSayW (Yell: PUnicodeChar); cdecl;
  begin
    DbgSay(PUnicodeCharToUnicodeString(Yell));
  end;

  procedure _ValidateTypesetsEq(n: PAnsiChar; v1, v2: integer); cdecl;
  begin
    try
    if v1 <> v2 then Die(MI_ERROR_TYPESETS_MISMATCH,[PCharToString(n), v1, v2]);
   except
    ProcessGuardedException();
   end;
  end;
  
  procedure _ValidateTypesetsLe(n: PAnsiChar; v1, v2: integer); cdecl;
  begin
    try
    if v1 > v2 then Die(MI_ERROR_TYPESETS_MISMATCH,[PCharToString(n), v1, v2]);
   except
    ProcessGuardedException();
   end;
  end;

  
  procedure _ValidateTypesets(i, n: integer); cdecl;
  begin
    try
    Case n of
      0: if i <> ord (MI_CORE_MIDs_END) then Die(MI_ERROR_TYPESETS_MISMATCH,['enumerated type TMessageId', i, ord (MI_CORE_MIDs_END)]);
      1: if i <> ord (Re_OnUnload) then Die(MI_ERROR_TYPESETS_MISMATCH,['enumerated type TCbMessage', i, ord (Re_OnUnload)]);
      5: if i <> sizeof(TMother) then Die(MI_ERROR_TYPESETS_MISMATCH,['size of interface variable Mother^', i, sizeof(TMother)]);
//      2: if i <> sizeof(TiCharArray) then Die(MI_ERROR_TYPESETS_MISMATCH,['TiCharArray', i, sizeof(TiCharArray)]);
    else
      Die(MI_ERROR_TYPESETS_MISMATCH,[PervertedFormat('ValidateTypesets(%0, %1)', [i, n]), 'n/a', 'n/a']);
    end;
   except
    ProcessGuardedException();
   end;
  end;

  procedure _RequestExit; cdecl;
  begin
    try
    if not(not Mother^.Display.CanRender or not Assigned(Console))
      then Console.SetDefaultBg(dbShutdown);
    Mother^.State.ExitRequested:=Yes;
   except
    ProcessGuardedException();
   end;
  end;

  procedure _ValidateVersion (m, n, b: integer); cdecl;
  begin
    try
    if (m > VersionMajor) or (n > VersionMinor) then
      DIE(MI_VERSIONS_MISMATCH, [VersionToStr(m, n, b), VersionToStr(VersionMajor, VersionMinor, BuildNumber)]);
   except
    ProcessGuardedException();
   end;
  end;

  procedure _cgeDrawBackground(); cdecl;
  begin
    try
     if Assigned(Console) then Console.DrawBackground();
   except
    ProcessGuardedException();
   end;
  end;
  
  threadvar _Mi2e: UnicodeString;

  function _HostMsg (M: TMessageId): PUnicodeChar; cdecl;
  begin
   try
     _Mi2e:=MessageContainer[M];
    Result:=@_Mi2e[1];
   except
    ProcessGuardedException();
   end;
  end;

  procedure FatalityMessageBox(w: PUnicodeChar); cdecl;
  begin
    {$ifdef windows}
      MessageBoxW(0, w, PUnicodeChar(RuEn(
        'УТРАЧЕНЫ ВСЕ ИЗМЕНЕНИЯ С МОМЕНТА ПРОШЛОГО СОХРАНЕНИЯ!',
        'ALL CHANGES MADE SINCE THE LAST SAVE WERE LOST!')), MB_ICONERROR or MB_SYSTEMMODAL or MB_OK);
    {$else}
      //***TODO: message box for Linux (use a framework?)
      WriteLn('All is lost! ', string(w));
    {$endif}
  end;

  procedure SetWindowName (n: PUnicodeChar); cdecl;
  begin
   try
    if Assigned(Framework)
      then Framework.SetWindowName(n);
   except
    ProcessGuardedException();
   end;
  end;

// *********************  TEXTURE MANAGEMENT ************************

  {$include cl_mother_api_assets.inc}

//*******************************************************************
  
  //These include implementation of
  //  procedure CgeBuildAlphaMipmaps(Width, Height: integer; Data: pointer); cdecl;
  {$ifdef ASSEMBLER_ALLOWED}
    {$include cle_alphamip_asm.inc}
  {$else}
    {$include cle_alphamip.inc}
  {$endif}
  

//******************************************************************  
