{$codepage utf-8}
{$longstrings on}
{$mode delphi} //NOT mode objfpc as the rest of the engine


unit cl_windows_missing;

{$define use_dynlibs}

interface
  uses Windows;

{$packrecords C}

{$ifdef cpu32}
  procedure LoadFunctionsUnsupportedInWindows98;
{$endif}

type
  TMonitorInfo = record
    cbSize: DWORD;
    rcMonitor: RECT;
    rcWork: RECT;
    dwFlags: DWORD;
  end;
  PMonitorInfo = ^TMonitorInfo;


const
  MONITOR_DEFAULTTONULL    = $00000000;
  MONITOR_DEFAULTTOPRIMARY = $00000001;
  MONITOR_DEFAULTTONEAREST = $00000002;

  function timeBeginPeriod(x1: UINT): MMRESULT; stdcall; external 'winmm.dll' name 'timeBeginPeriod';
  function timeEndPeriod(x1: UINT): MMRESULT; stdcall;  external 'winmm.dll' name 'timeEndPeriod';

type
  MMRESULT = UINT;
const
  TIMERR_NOERROR = 0;
  TIMERR_NOCANDO = 96 + 1;


{$ifdef cpu64}
  //static linking: 64 bits means Windows XP or higher.
  function MonitorFromRect (const lprc: RECT; dwFlags: DWORD): HMONITOR;
                                                         external 'user32.dll';

  function GetMonitorInfo (hMonitor: HMONITOR; lpmi: PMonitorInfo): BOOL;
                                  external 'user32.dll' name 'GetMonitorInfoA';

  function CreateWindowExW(dwExStyle:DWORD; lpClassName:LPCWSTR;
    lpWindowName:LPCWSTR; dwStyle:DWORD; X:longint;Y:longint;
    nWidth:longint; nHeight:longint;
    hWndParent:HWND; hMenu:HMENU;hInstance:HINST; lpParam:LPVOID):HWND;
                                                         external 'user32.dll';

  function SetWindowTextW(hWnd:HWND; lpString:LPCWSTR):WINBOOL;
                                                         external 'user32.dll';

  function TerminateProcess (hProcess:HANDLE; uExitCode:UINT):WINBOOL;
                                                       external 'kernel32.dll';

  function SHGetFolderPathA  (HWND:hwnd;csidl:dword;htoken:THandle;
                                      dwflags:dword;pszpath:PAnsiChar):HResult;
                                                        external 'shell32.dll';

{$else cpu64}
  //32 bits: must not link statically,
  //  to avoid losing compatibility with vanilla Windows 98 SE!

var
  MonitorFromRect: function (const lprc: RECT; dwFlags: DWORD): HMONITOR; stdcall;
  GetMonitorInfo: function (hMonitor: HMONITOR; lpmi: PMonitorInfo): BOOL; stdcall;

  TerminateProcess: function(hProcess:HANDLE; uExitCode:UINT):WINBOOL; stdcall;


  SHGetFolderPathA: function
//    (p1: pointer; p2: integer; p3: pointer; p4: dword; p5: pointer): HResult; stdcall;
    (HWND:hwnd;csidl:dword;htoken:THandle;dwflags:dword;pszpath:PAnsiChar):HResult; StdCall;

  CreateWindowExW: function(dwExStyle:DWORD; lpClassName:LPCWSTR;
    lpWindowName:LPCWSTR; dwStyle:DWORD; X:longint;Y:longint;
    nWidth:longint; nHeight:longint;
    hWndParent:HWND; hMenu:HMENU;hInstance:HINST; lpParam:LPVOID):HWND; stdcall;

  SetWindowTextW: function(hWnd:HWND; lpString:LPCWSTR):WINBOOL; stdcall;
{$endif cpu64}

  const
    SHGFP_TYPE_CURRENT = 0;
    CSIDL_PROGRAM_FILES = $26;
    CSIDL_LOCAL_APPDATA = $1c;
    CSIDL_APPDATA = $1a;
    CSIDL_FLAG_NO_ALIAS = $1000;
    CSIDL_FLAG_CREATE = $8000; { (force creation of requested folder if it doesn't exist yet)     }

implementation
{$ifdef cpu32}
  uses cge, chelinfo;

  function LoadProcAddress(hdll: THandle; ProcName: ansistring) : pointer;
  var
    baseaddr: pointer;
    exename: UnicodeString;
  begin
    Result:= GetProcAddress(hdll, PChar(ProcName));
    if Assigned(Result) then begin
      GetModuleByAddr(Result, baseaddr, exename);
      VerboseLog('    %0() at %1 in %2', [ProcName, Result, exename]);
    end
    else
      VerboseLog('    %0() not found!', [ProcName]);
  end;

  procedure LoadFunctionsUnsupportedInWindows98;
  var
    shell32, kernel32, user32: THandle;
  begin

    //lol this is called when the log and config aren't yet initialized

    //use *cached* verbose log (i.e. it is cached anyway but discarded later
    //  if there is no verbose option set in the config
    VerboseLog('Dynamically loading Win32 API functions to ensure compatibility with vanilla Windows 98 SE...');

    VerboseLog('  loading shell32.dll...');
    shell32:= LoadLibrary('shell32.dll');

    @SHGetFolderPathA:= LoadProcAddress(shell32, 'SHGetFolderPathA');
    Mother^.State.Windows98.DoesNotSupportUserDirectory:= not (
      Assigned(SHGetFolderPathA)
    );

//    FreeLibrary(shell32);

    VerboseLog('  loading kernel32.dll...');
    kernel32:= LoadLibrary('kernel32.dll');

    @TerminateProcess:= LoadProcAddress(kernel32, 'TerminateProcess');

//    FreeLibrary(kernel32);

    VerboseLog('  loading user32.dll...');
    user32:= LoadLibrary('user32.dll');

    @MonitorFromRect:= LoadProcAddress(user32, 'MonitorFromRect');
    @GetMonitorInfo:= LoadProcAddress(user32, 'GetMonitorInfoA');
    @CreateWindowExW:= LoadProcAddress(user32, 'CreateWindowExW');
    @SetWindowTextW:= LoadProcAddress(user32, 'SetWindowTextW');

//    pointer():= LoadProcAddress(user32, '');
//    pointer():= LoadProcAddress(user32, '');
//    pointer():= LoadProcAddress(user32, '');

//    FreeLibrary(user32);

  end;
{$endif}
end.
