

  procedure RetrieveAssets;
  var
    s: TStream;
    h: pointer;
    i: integer;
  begin
    if Mother^.Debug.Verbose
      then AddLog('  Retrieving assets from the mother memory stream...');
    if Assigned(gv_RetrievedAssets) then begin
//***TODO: what if it is still non-empty when the session is loaded *again* ?
//Answer: before it is loaded, asset list must be exported anyway.
//Any assets from gv_RetrievedAssets are moved to gv_AssetsToStore

      gv_RetrievedAssets.ClearAndFree;
      gv_RetrievedAssets.Free;
    end;
    pointer(gv_RetrievedAssets):= NIL;
    h:= Mother^.Core.GetAssetStorageMemoryStream();
    if not Assigned(h) then Exit;
    s:= TMotherStream.CreateFromHandle(h, false);
     //IDIOT, the mother ci stream becomes invalid! //true); //delete mother stream as well when done!
    s.Position:= 0;
//bgSay('* stream size='+IntToStr(s.Size)+ ', position=' + IntToStr(s.position));
    gv_RetrievingFromMother:= true;
    gv_RetrievedAssets:= CpsLoad( RuEn('списка ресурсов', 'assets list'), s) as TAssetContainer;
    gv_RetrievingFromMother:= false;
    s.Free;
    if Mother^.Debug.Verbose then begin
      if Assigned(gv_RetrievedAssets)
        then begin
          AddLog('  Ok, total %0', [Length(gv_RetrievedAssets.List)]);
          for i:= 0 to High(gv_RetrievedAssets.List) do
            if gv_RetrievedAssets.List[i].Exists
              then AddLog('    #%0: %1, hash %2', [i, string(gv_RetrievedAssets.List[i].ClassName), HashToHex(gv_RetrievedAssets.List[i].Hash)])
              else AddLog('    #%0: invalid (%1)', [i, TellInvalidAsset(gv_RetrievedAssets.List[i])]);
        end
        else AddLog('  Ok, the loaded list is NIL.')
    end;
    Mother^.Core.SetAssetStorageMemoryStream(NIL);
    Mother^.Stream.Free(h);
  end;


  procedure StoreAssets;
  var
    s: TStream;
    h: pointer;
    i: integer;
  begin
    //the main reason for these shamanistic dances is the mother and the module
    //  having two independent memory managers
    if Mother^.Debug.Verbose
      then AddLog('  Storing assets to a mother memory stream...');

    if not Assigned(gv_AssetsToStore) then gv_AssetsToStore:= TAssetContainer.Create;

    gv_StoringToMother:= true;
    with gv_AssetsToStore do begin
      MergeAndFree(gv_RetrievedAssets);
        { because dealing with unclaimed assets is *lazy*, there could be
          a situation when the session is loaded again before they are all
          neutralized.
         We just pass them further on by moving them to the list. }
      MergeAndFree(gv_TemporaryAssets);
      MassSetUnclaimed;
      MassResurrectScraped; { if there are some (unlikely) both actualized
        and scraped, the garbage collector will free them when the module
        goes down, which would invalidate the stored handles and such }
      ZeroAllNotActualized;
      Remove(nil);

      h:= Mother^.Core.GetAssetStorageMemoryStream();
      { For the mother, h is just a TMemoryStream cast to pointer.
        But for the module, it's a non-transparent handle, since the module DLL
          has its own copies of the RTL classes, which are, of course,
          not compatible with those of the mother. }

      if Assigned(h) then Mother^.Stream.Free(h); { because there's no #$& way
                                         to just reset its size back to zero }
{
Question: How did the mother memory stream get set to an invalid value?

There were unprocessed error messages left:
!!..
Program crashed with an uncaught exception.
The guarded function exception
  re-raised at cl_resource_func.inc:428 in chentrah-debug.exe
EAccessViolation: Access violation
  at 00408D0Fh in chentrah-debug.exe
  (no debug info available:
doesn't include this address range)
Call stack:
  cl_resource_func.inc:428 in chentrah-debug.exe
  mo_assets_stor.inc:69 in _test018.dll
  mo_module.pp:831 in _test018.dll
  mo_module.pp:1124 in _test018.dll
  cl_module.inc:455 in chentrah-debug.exe
  cl_window.inc:286 in chentrah-debug.exe
  cl_winman.inc:385 in chentrah-debug.exe
  cl_framework_winapi.pp:138 in chentrah-debug.exe
  cl_main.inc:439 in chentrah-debug.exe
  chentrah.lpr:115 in chentrah-debug.exe
  0040EB61h in chentrah-debug.exe

}

      h:= Mother^.Memory.CreateStream();
      Mother^.Core.SetAssetStorageMemoryStream(h);
      s:= TMotherStream.CreateFromHandle(h, false);

      if not CpsStore(
        gv_AssetsToStore, RuEn('списка ресурсов', 'assets list'),
        s, $00000000, ASSET_MASK)
      then begin
        s.Free;
        Mother^.Core.SetAssetStorageMemoryStream(NIL);
        Mother^.Stream.Free(h);
        Mother^.State.StateTrashedRestartRequired:= Yes;
        gv_StoringToMother:= false;
        Die(GetChepersyErrorLog());
      end;

      s.Free; //does NOT free the mother stream as we passed false to the constructor

      if Mother^.Debug.Verbose then AddLog('  Ok, total %0'
        , [Length(List)]);
    end;
    gv_StoringToMother:= false;
  end;


  procedure DealWithUnclaimedAssets;
  var
    q: qword;
    i, t, c : integer;
  begin
    //For *now*, just delete them at once, assuming that the cost of
    //  freeing whatever is negligible.
    //***TODO: benchmark and optimize the shit out of this if needed.

    if not Assigned(gv_RetrievedAssets) then Exit;

    t:= Length(gv_RetrievedAssets.List);

    if (Mother^.Debug.Verbose) then AddLog('  Deleting %0 unclaimed assets...', [t]);
    Mother^.Timer.UsecByTsc(@q);

    c:=0;
    for i:= 0 to High(gv_RetrievedAssets.List) do
      with gv_RetrievedAssets.List[i] do begin
        if not Exists then continue;
        if Actualized
          then begin
            if (Mother^.Debug.Verbose) then AddLog('    ..deleting %0', [string(ClassName)]);
            Free;
            c+= 1;
          end;
      end;

    if (Mother^.Debug.Verbose) then AddLog('  Deleted %0 unclaimed assets in %1µs',[c, round(Mother^.Timer.UsecByTsc(@q))]);
    gv_RetrievedAssets.Free;
    pointer(gv_RetrievedAssets):= nil;
  end;


  procedure FindAssetsWalkProc(o: TManagedObject);
  begin
    if o.Exists()
      and ((o.CpsMask and ASSET_MASK) <> 0)
      and (o is TGenericAsset)
      and (o as TGenericAsset).Actualized
    then begin
      gv_AssetsToStore.Add(o as TGenericAsset);
    end;
  end;

  procedure WalkSessionToListAssetsToStore;
    { Is used when not really saving the session (like in the case of a
        rollback). Otherwise the saving routine itself serves this role. }
  begin
    if Mother^.Debug.Verbose then AddLog('  Parsing the data tree to mark assets for storing...');
    if not CpsWalkGraph(Module.Logic, @FindAssetsWalkProc)
    then begin
      Mother^.State.StateTrashedRestartRequired:= Yes;
      Die(GetChepersyErrorLog());
    end;
    if Mother^.Debug.Verbose then AddLog('  Ok, found %0.', [Length(gv_AssetsToStore.List)]);
  end;


  constructor TAssetContainer.Create;
  begin
    inherited Create;
    Self.CpsMask:= Self.CpsMask or ASSET_MASK;
  end;

  procedure TAssetContainer.Add(A: TGenericAsset);
  begin
    SetLength(List, Length(List) + 1);
    List[High(List)]:= A;
  end;


  destructor TAssetContainer.Destroy;
  begin
    SetLength(List, 0);
    inherited;
  end;

  procedure TAssetContainer.ClearAndFree;
  var A: TGenericAsset;
  begin
    for A in List do
      if Assigned(A) and not A.Scraped then A.Scrape;
    SetLength(List, 0);
  end;

  procedure TAssetContainer.ClearAndForget;
  begin
    SetLength(List, 0);
  end;


  function TAssetContainer.GetByHash(ph: PAssetHash): TGenericAsset;
  var i: integer;
  begin
    for i:= High(List) downto 0 {backwards order is more natural,
      according to how they are stored and how AfterLoading is called}
    do begin
      if not Assigned(List[i]) then continue; {Chepersy *may* return an array
                             with some elements discarded and filled with NIL }
      if ap_temporary in List[i].AssetProperties then continue; { temporary
                                          assets should be deleted regardless }
      if ph^ = List[i].Hash then begin
        Result:= List[i];
        if i < High(List) then List[i]:= List[High(List)];
        SetLength(List, Length(List) - 1);
        Exit;
      end;
    end;
    Result:= NIL;
  end;

  procedure TAssetContainer.Remove(a: TGenericAsset);
  var i, nh: integer;
  begin
    i:=0;
    nh:= High(List);
    while i < nh do begin
      if List[i] = a then begin
        if i < nh then List[i]:= List[nh];
        dec(nh);
      end;
      inc(i);
    end;
    SetLength(List, nh + 1);
  end;

  procedure TAssetContainer.MergeAndFree(var c2: TAssetContainer); //free and nil
  var i, j: integer;
  begin
    if not Assigned(c2) then Exit;
    j:= Length(List);
    SetLength(List, j + Length(c2.List));
    for i:= 0 to High(c2.List) do List[j + i]:= c2.List[i];
    c2.Free;
    pointer(c2):= nil;
  end;

  procedure TAssetContainer.MassSetUnclaimed;
  var i: integer;
  begin
    for i:=0 to High(List) do if Assigned(List[i]) then List[i].MarkAsUnclaimed;
  end;

  procedure TAssetContainer.MassResurrectScraped;
  var i: integer;
  begin
    for i:=0 to High(List) do if Assigned(List[i])
      and List[i].Scraped and List[i].Actualized
    then begin
      List[i].Resurrect; //pull out of the garbage collector graveyard
      List[i].MarkAsTemporary; //will lead to scraping in the next reincarnation
    end;
  end;

  procedure TAssetContainer.ZeroAllNotActualized;
  var i: integer;
  begin
    for i:=0 to High(List) do if Assigned(List[i]) and not List[i].Actualized
      then pointer(List[i]):= nil;
  end;

  procedure TAssetContainer.RegisterFields;
  begin
    RegType(typeinfo(TAssetArray), typeinfo(TGenericAsset));
    inherited;
    ListFields([
      'list', @list, typeinfo(TAssetArray)
    ]);
  end;

