{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{
    Note: at first, images were assets. But then I realized that they were
    too volatile for that, too temporary. Unnneeded after they are loaded
    into that texture or whatever.

    So I made images into threaded tasks.
    They live while the  owner asset actualizes. Then they die.

    On a side note, I should work on moving from Vampyre Imaging library
    to SDL2 image or something, and incorporate imaging into the framework.
    Because, as awesome as Vampyre is, I don't see it actually update since 2009.

    Note 2: this operates with opengl texture format constants, but the
    imaging library uses its own, different format representation.
    Format designations are converted on the fly.
}

unit mo_image;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, md5,  chepersy, mo_hub, typinfo, mo_classes,
    mo_gmathbase, mo_threads;


type

  TImageDecodeTask = class(TAbstractThreadedTask)
  private
    f_pinbuff, f_poutbuff: pointer;
    f_insize, f_outsize: integer;
    f_imghandle: pointer; //in mother, this resolves to a class instance
    f_width, f_height: integer;
    f_format: GLuint;
    function Perform: boolean; override;
  public
    property Width: integer read f_width;
    property Height: integer read f_height;
    property Format: GLuint read f_format;
    property ImageHandle: pointer read f_imghandle;
    property PPixels: pointer read f_poutbuff;
    property PixelsSize: integer read f_outsize;
    function GetType: TThreadedTaskType; override;

    constructor Create(pbuf: pointer; bsize: longint);
    destructor Destroy; override;
  end;

  TImageResampleTask = class(TImageDecodeTask)
  private
    f_source: TImageDecodeTask;
    f_resample_width,
    f_resample_height : integer;
    f_resample_format: GLuint;
    f_cycled: boolean;
    function Perform: boolean; override;
  public
    property Source: TImageDecodeTask read f_source;

    constructor Create(src: TImageDecodeTask; newx, newy: integer;
      newformat: GLuint; cycled: boolean = false);
  end;

{

***TODO
  TImageEncodeTask = class(TAbstractThreadedTask)
  private

  public
  end;
}



implementation

  function TImageDecodeTask.Perform: boolean;
  var
    q: qword;
  begin
    Result:= true; //in case of failure, consider task completed as well
    if Mother^.Debug.Verbose then AddLog('TImageDecodeTask.Perform');
    if Mother^.Debug.Verbose then AddLog('  encoded size=%0', [ f_insize]);
    if not Assigned(f_pinbuff) or (f_insize <= 0)
      then Die(RuEn(
        'Попытка декодировать изображение из пустого буфера!',
        'Attempt to decode image from empty buffer!'));
//bgSay('doing byte(nil^):= 0 ...');
//byte(nil^):= 0;
    Mother^.Timer.UsecbyTsc(@q);
    f_imghandle:= Mother^.Image.LoadFromMem(f_pinbuff, f_insize);
    AfterEfCheck;
    f_poutbuff:= Mother^.Image.GetPixels(f_imghandle);
    AfterEfCheck;
    f_outsize:= Mother^.Image.GetBytes(f_imghandle);
    AfterEfCheck;
    f_width:= Mother^.Image.GetWidth(f_imghandle);
    AfterEfCheck;
    f_height:= Mother^.Image.GetHeight(f_imghandle);
    AfterEfCheck;
    f_format:= Mother^.Image.GetFmt(f_imghandle);
    AfterEfCheck;
    if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecByTsc(@q))]);
    if Mother^.Debug.Verbose then AddLog('  dimensions %0x%1, format %2h', [f_width, f_height, GlTextureFormatToString(f_format)]);
  end;

  function TImageResampleTask.Perform: boolean;
  var
    q: qword;
  begin
    if not f_source.Completed then Exit(false); //postpone. This is unlikely situation, but still.

    Result:= true; //in case of failure, consider task completed as well
    if Mother^.Debug.Verbose then AddLog('TImageResampleTask.Perform');
    if Mother^.Debug.Verbose then AddLog('  source dimensions %0x%1, format %2h', [f_source.Width, f_source.Height, GlTextureFormatToString(f_source.Format)]);
    Mother^.Timer.UsecbyTsc(@q);
    if (f_resample_width = f_source.Width) and (f_resample_height = f_source.Height)
    then begin
      //sizes equal, only format differs.
      //in fact, I don't want to know if resample routine is smart enough.
      //we just convert the format;
      f_imghandle:= Mother^.Image.Clone(f_source.ImageHandle);
      AfterEfCheck;
      Mother^.Image.SetFmt(f_imghandle, f_resample_format);
      AfterEfCheck;
      if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecByTsc(@q))]);
      if Mother^.Debug.Verbose then AddLog('  format %0h', [GlTextureFormatToString(f_format)]);
    end
    else begin
      f_imghandle:= Mother^.Image.CloneResample (f_source.ImageHandle,
        f_resample_width, f_resample_height, f_cycled, f_resample_format);
      AfterEfCheck;
      f_width:= Mother^.Image.GetWidth(f_imghandle);
      AfterEfCheck;
      f_height:= Mother^.Image.GetHeight(f_imghandle);
      AfterEfCheck;
      f_format:= Mother^.Image.GetFmt(f_imghandle);
      AfterEfCheck;
      if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecByTsc(@q))]);
      if Mother^.Debug.Verbose then AddLog('  dimensions %0x%1, format %2h', [f_width, f_height, GlTextureFormatToString(f_format)]);
    end;
  end;

  constructor TImageDecodeTask.Create(pbuf: pointer; bsize: longint);
  begin
    inherited Create;
    f_pinbuff:= pbuf;
    f_insize:= bsize;
    Reschedule;
  end;

  constructor TImageResampleTask.Create(src: TImageDecodeTask;
    newx, newy: integer;  newformat: GLuint; cycled: boolean = false);
  begin
    f_source:= src;
    f_resample_width:= newx;
    f_resample_height:= newy;
    f_resample_format:= newformat;
    f_cycled:= cycled;
    Reschedule;
  end;

  destructor TImageDecodeTask.Destroy;
  begin
    if Assigned(f_imghandle) then begin
      Mother^.Image.Delete(f_imghandle);
      AfterEfCheck;
    end;
    inherited;
  end;

  function TImageDecodeTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_Imaging;
  end;

end.
